#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2017--2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
SacreBLEU provides hassle-free computation of shareable, comparable, and reproducible BLEU scores.
Inspired by Rico Sennrich's `multi-bleu-detok.perl`, it produces the official WMT scores but works with plain text.
It also knows all the standard test sets and handles downloading, processing, and tokenization for you.

See the [README.md] file for more information.
"""
from .sacrebleu import main

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

import os
import logging
import urllib.request

from functools import lru_cache
from ..utils import SACREBLEU_DIR, download_file
from .tokenizer_base import BaseTokenizer

sacrelogger = logging.getLogger('sacrebleu')


class TokenizerSPM(BaseTokenizer):
    def signature(self):
        return 'spm-flores'

    def __init__(self):
        try:
            import sentencepiece as spm
        except (ImportError, ModuleNotFoundError):
            raise ImportError(
                '\n\nPlease install the sentencepiece library for SPM tokenization:'
                '\n\n  pip install sentencepiece '
            )
        self.sp = spm.SentencePieceProcessor()

        tokenizer_path = os.path.join(SACREBLEU_DIR, "models", "flores_sacrebleu_tokenizer_spm.model")
        if not os.path.exists(tokenizer_path):
            url = "https://dl.fbaipublicfiles.com/fairseq/models/flores/sacrebleu_tokenizer_spm.model"
            download_file(url, tokenizer_path)
        self.sp.Load(tokenizer_path)

    @lru_cache(maxsize=None)
    def __call__(self, line):
        """Tokenizes all the characters in the input line.

        :param line: a segment to tokenize
        :return: the tokenized line
        """
        return " ".join(self.sp.EncodeAsPieces(line))



class BaseTokenizer:
    """A base dummy tokenizer to derive from."""

    def signature(self):
        """
        Returns a signature for the tokenizer.

        :return: signature string
        """
        return 'none'

    def __call__(self, line):
        """
        Tokenizes an input line with the tokenizer.

        :param line: a segment to tokenize
        :return: the tokenized line
        """
        return line



import pytest
import sacrebleu

EPSILON = 1e-3


# Example taken from #98
REF = "producţia de zahăr brut se exprimă în zahăr alb;"
SYS = "Producția de zahăr primă va fi exprimată în ceea ce privește zahărul alb;"

test_cases = [
    # change smoothing
    ('exp', None, False, '13a', 8.493),
    ('none', None, False, '13a', 0.0),
    ('floor', None, False, '13a', 4.51688),    # defaults to 0.1
    ('floor', 0.1, False, '13a', 4.51688),
    ('floor', 0.5, False, '13a', 10.10),
    ('add-k', None, False, '13a', 14.882),     # defaults to 1
    ('add-k', 1, False, '13a', 14.882),
    ('add-k', 2, False, '13a', 21.389),
    # change tok
    ('exp', None, False, 'none', 7.347),
    ('exp', None, False, 'intl', 8.493),
    ('exp', None, False, 'char', 40.8759),
    # change case
    ('exp', None, True, 'char', 42.0267),
]


# Example taken from #141
REF_0 = "okay thanks"
SYS_0 = "this is a cat"

test_cases_zero_bleu = [
    ('exp', None, False, '13a', 0.0),
    ('none', None, False, '13a', 0.0),
    ('floor', None, False, '13a', 0.0),    # defaults to 0.1
    ('floor', 0.1, False, '13a', 0.0),
    ('add-k', None, False, '13a', 0.0),     # defaults to 1
    ('add-k', 1, False, '13a', 0.0),
]


@pytest.mark.parametrize("smooth_method, smooth_value, lowercase, tok, expected_score", test_cases)
def test_compat_sentence_bleu(smooth_method, smooth_value, lowercase, tok, expected_score):
    score = sacrebleu.compat.sentence_bleu(
        SYS, [REF], smooth_method=smooth_method, smooth_value=smooth_value,
        tokenize=tok,
        lowercase=lowercase,
        use_effective_order=True)
    assert abs(score.score - expected_score) < EPSILON


@pytest.mark.parametrize("smooth_method, smooth_value, lowercase, tok, expected_score", test_cases)
def test_api_sentence_bleu(smooth_method, smooth_value, lowercase, tok, expected_score):
    metric = sacrebleu.metrics.BLEU(
        lowercase=lowercase, force=False, tokenize=tok,
        smooth_method=smooth_method, smooth_value=smooth_value,
        effective_order=True)
    score = metric.sentence_score(SYS, [REF])

    assert abs(score.score - expected_score) < EPSILON


@pytest.mark.parametrize("smooth_method, smooth_value, lowercase, tok, expected_score", test_cases_zero_bleu)
def test_api_sentence_bleu_zero(smooth_method, smooth_value, lowercase, tok, expected_score):
    metric = sacrebleu.metrics.BLEU(
        lowercase=lowercase, force=False, tokenize=tok,
        smooth_method=smooth_method, smooth_value=smooth_value,
        effective_order=True)
    score = metric.sentence_score(SYS_0, [REF_0])
    assert abs(score.score - expected_score) < EPSILON



# Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pytest
import sacrebleu

EPSILON = 1e-4

test_sentence_level_chrf = [
    (
        'Co nás nejvíc trápí, protože lékaři si vybírají, kdo bude žít a kdo zemře.',
        ['Nejvíce smutní jsme z toho, že musíme rozhodovat o tom, kdo bude žít a kdo zemře.'],
        39.14078509,
    ),
    (
        'Nebo prostě nemají vybavení, které by jim pomohlo, uvedli lékaři.',
        ['A někdy nemáme ani potřebný materiál, abychom jim pomohli, popsali lékaři.'],
        31.22557079,
    ),
    (
        'Lapali po dechu, jejich životy skončily dřív, než skutečně začaly.',
        ['Lapali po dechu a pak jejich život skončil - dřív, než skutečně mohl začít, připomněli.'],
        57.15704367,
    ),
]


# hypothesis, reference, expected score
# >= 2.0.0: some orders are not fulfilled in epsilon smoothing (chrF++.py and NLTK)
test_cases = [
    (["abcdefg"], ["hijklmnop"], 0.0),
    (["a"], ["b"], 0.0),
    ([""], ["b"], 0.0),
    ([""], ["ref"], 0.0),
    ([""], ["reference"], 0.0),
    (["aa"], ["ab"], 8.3333),
    (["a", "b"], ["a", "c"], 8.3333),
    (["a"], ["a"], 16.6667),
    (["a b c"], ["a b c"], 50.0),
    (["a b c"], ["abc"], 50.0),
    ([" risk assessment must be made of those who are qualified and expertise in the sector - these are the scientists ."],
     ["risk assessment has to be undertaken by those who are qualified and expert in that area - that is the scientists ."], 63.361730),
    ([" Die    Beziehung zwischen  Obama und Netanjahu ist nicht gerade  freundlich. "],
     ["Das Verhältnis zwischen Obama und Netanyahu ist nicht gerade freundschaftlich."], 64.1302698),
    (["Niemand hat die Absicht, eine Mauer zu errichten"], ["Niemand hat die Absicht, eine Mauer zu errichten"], 100.0),
]

# sacreBLEU < 2.0.0 mode
# hypothesis, reference, expected score
test_cases_effective_order = [
    (["a"], ["a"], 100.0),
    ([""], ["reference"], 0.0),
    (["a b c"], ["a b c"], 100.0),
    (["a b c"], ["abc"], 100.0),
    ([""], ["c"], 0.0),
    (["a", "b"], ["a", "c"], 50.0),
    (["aa"], ["ab"], 25.0),
]

test_cases_keep_whitespace = [
    (
        ["Die Beziehung zwischen Obama und Netanjahu ist nicht gerade freundlich."],
        ["Das Verhältnis zwischen Obama und Netanyahu ist nicht gerade freundschaftlich."],
        67.3481606,
    ),
    (
        ["risk assessment must be made of those who are qualified and expertise in the sector - these are the scientists ."],
        ["risk assessment has to be undertaken by those who are qualified and expert in that area - that is the scientists ."],
        65.2414427,
    ),
]


@pytest.mark.parametrize("hypotheses, references, expected_score", test_cases)
def test_chrf(hypotheses, references, expected_score):
    score = sacrebleu.corpus_chrf(
        hypotheses, [references], char_order=6, word_order=0, beta=3,
        eps_smoothing=True).score
    assert abs(score - expected_score) < EPSILON


@pytest.mark.parametrize("hypotheses, references, expected_score", test_cases_effective_order)
def test_chrf_eff_order(hypotheses, references, expected_score):
    score = sacrebleu.corpus_chrf(
        hypotheses, [references], char_order=6, word_order=0, beta=3,
        eps_smoothing=False).score
    assert abs(score - expected_score) < EPSILON


@pytest.mark.parametrize("hypotheses, references, expected_score", test_cases_keep_whitespace)
def test_chrf_keep_whitespace(hypotheses, references, expected_score):
    score = sacrebleu.corpus_chrf(
        hypotheses, [references], char_order=6, word_order=0, beta=3,
        remove_whitespace=False).score
    assert abs(score - expected_score) < EPSILON


@pytest.mark.parametrize("hypothesis, references, expected_score", test_sentence_level_chrf)
def test_chrf_sentence_level(hypothesis, references, expected_score):
    score = sacrebleu.sentence_chrf(hypothesis, references, eps_smoothing=True).score
    assert abs(score - expected_score) < EPSILON



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2017--2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

__version__ = '2.1.0'
__description__ = 'Hassle-free computation of shareable, comparable, and reproducible BLEU, chrF, and TER scores'


from .utils import smart_open, SACREBLEU_DIR, download_test_set  # noqa: F401
from .utils import get_source_file, get_reference_files  # noqa: F401
from .utils import get_available_testsets, get_langpairs_for_testset  # noqa: F401
from .metrics.helpers import extract_word_ngrams, extract_char_ngrams  # noqa: F401
from .dataset import DATASETS  # noqa: F401
from .metrics import BLEU, CHRF, TER  # noqa: F401

# Backward compatibility functions for old style API access (<= 1.4.10)
from .compat import corpus_bleu, raw_corpus_bleu, sentence_bleu  # noqa: F401
from .compat import corpus_chrf, sentence_chrf  # noqa: F401
from .compat import corpus_ter, sentence_ter  # noqa: F401



import os

from collections import defaultdict

from sacrebleu.metrics import BLEU
from sacrebleu.significance import PairedTest

import pytest


def _read_pickle_file():
    import bz2
    import pickle as pkl
    with bz2.BZ2File('./test/wmt17_en_de_systems.pkl.bz2', 'rb') as f:
        data = pkl.load(f)
    return data


# P-values obtained from Moses' significance script (mean of 3 runs)
# Script: scripts/moses-sigdiff.pl (modified to bootstrap samples = 2000)
MOSES_P_VALS = {
    "newstest2017.C-3MA.4959.en-de": 0.00000,
    "newstest2017.FBK.4870.en-de": 0.01267,
    "newstest2017.KIT.4950.en-de": 0.02233,
    "newstest2017.LMU-nmt-reranked.4934.en-de": 0.04383,
    "newstest2017.LMU-nmt-single.4893.en-de": 0.20783,
    "newstest2017.online-A.0.en-de": 0.00000,
    "newstest2017.online-B.0.en-de": 0.38100,
    "newstest2017.online-F.0.en-de": 0.00000,
    "newstest2017.online-G.0.en-de": 0.00000,
    "newstest2017.PROMT-Rule-based.4735.en-de": 0.00000,
    "newstest2017.RWTH-nmt-ensemble.4921.en-de": 0.01167,
    "newstest2017.SYSTRAN.4847.en-de": 0.20983,
    "newstest2017.TALP-UPC.4834.en-de": 0.00000,
    "newstest2017.uedin-nmt.4722.en-de": 0.00000,
    "newstest2017.xmu.4910.en-de": 0.25483,
}

# Obtained from the multeval toolkit, 10,000 AR trials, (BLEU and TER)
# Code: github.com/mjclark/multeval.git
MULTEVAL_P_VALS = {
    "newstest2017.C-3MA.4959.en-de": (0.0001, 0.0001),
    "newstest2017.FBK.4870.en-de": (0.0218, 0.09569),
    "newstest2017.KIT.4950.en-de": (0.0410, 0.0002),
    "newstest2017.LMU-nmt-reranked.4934.en-de": (0.09029, 0.0001),
    "newstest2017.LMU-nmt-single.4893.en-de": (0.58494, 0.0054),
    "newstest2017.online-A.0.en-de": (0.0001, 0.0001),
    "newstest2017.online-B.0.en-de": (0.94111, 0.82242),
    "newstest2017.online-F.0.en-de": (0.0001, 0.0001),
    "newstest2017.online-G.0.en-de": (0.0001, 0.0001),
    "newstest2017.PROMT-Rule-based.4735.en-de": (0.0001, 0.0001),
    "newstest2017.RWTH-nmt-ensemble.4921.en-de": (0.0207, 0.07539),
    "newstest2017.SYSTRAN.4847.en-de": (0.59914, 0.0001),
    "newstest2017.TALP-UPC.4834.en-de": (0.0001, 0.0001),
    "newstest2017.uedin-nmt.4722.en-de": (0.0001, 0.0001),
    "newstest2017.xmu.4910.en-de": (0.71073, 0.0001),
}


SACREBLEU_BS_P_VALS = defaultdict(float)
SACREBLEU_AR_P_VALS = defaultdict(float)

# Load data from pickled file to not bother with WMT17 downloading
named_systems = _read_pickle_file()
_, refs = named_systems.pop()
metrics = {'BLEU': BLEU(references=refs, tokenize='none')}


#########
# BS test
#########
os.environ['SACREBLEU_SEED'] = str(12345)
bs_scores = PairedTest(
    named_systems, metrics, references=None,
    test_type='bs', n_samples=2000)()[1]

for name, result in zip(bs_scores['System'], bs_scores['BLEU']):
    if result.p_value is not None:
        SACREBLEU_BS_P_VALS[name] += result.p_value


###############################################
# AR test (1 run)
# Test only BLEU as TER will take too much time
###############################################
ar_scores = PairedTest(named_systems, metrics, references=None,
                       test_type='ar', n_samples=10000)()[1]

for name, result in zip(ar_scores['System'], ar_scores['BLEU']):
    if result.p_value is not None:
        SACREBLEU_AR_P_VALS[name] += result.p_value


@pytest.mark.parametrize("name, expected_p_val", MOSES_P_VALS.items())
def test_paired_bootstrap(name, expected_p_val):
    p_val = SACREBLEU_BS_P_VALS[name]
    assert abs(p_val - expected_p_val) < 1e-2


@pytest.mark.parametrize("name, expected_p_vals", MULTEVAL_P_VALS.items())
def test_paired_approximate_randomization(name, expected_p_vals):
    expected_bleu_p_val = expected_p_vals[0]
    p_val = SACREBLEU_AR_P_VALS[name]
    assert abs(p_val - expected_bleu_p_val) < 1e-2



from functools import lru_cache
from .tokenizer_base import BaseTokenizer
from .tokenizer_re import TokenizerRegexp


class Tokenizer13a(BaseTokenizer):

    def signature(self):
        return '13a'

    def __init__(self):
        self._post_tokenizer = TokenizerRegexp()

    @lru_cache(maxsize=2**16)
    def __call__(self, line):
        """Tokenizes an input line using a relatively minimal tokenization
        that is however equivalent to mteval-v13a, used by WMT.

        :param line: a segment to tokenize
        :return: the tokenized line
        """

        # language-independent part:
        line = line.replace('<skipped>', '')
        line = line.replace('-\n', '')
        line = line.replace('\n', ' ')

        if '&' in line:
            line = line.replace('&quot;', '"')
            line = line.replace('&amp;', '&')
            line = line.replace('&lt;', '<')
            line = line.replace('&gt;', '>')

        return self._post_tokenizer(f' {line} ')



"""The implementation of chrF (Popović 2015) and chrF++ (Popović 2017) metrics."""

from typing import List, Sequence, Optional, Dict
from collections import Counter

from ..utils import sum_of_lists
from .base import Score, Signature, Metric
from .helpers import extract_all_char_ngrams, extract_word_ngrams


class CHRFSignature(Signature):
    """A convenience class to represent the reproducibility signature for chrF.

    :param args: key-value dictionary passed from the actual metric instance.
    """
    def __init__(self, args: dict):
        """`CHRFSignature` initializer."""
        super().__init__(args)
        self._abbr.update({
            'case': 'c',
            'eff': 'e',
            'nc': 'nc',
            'nw': 'nw',
            'space': 's',
        })

        self.info.update({
            'case': 'lc' if args['lowercase'] else 'mixed',
            'eff': 'yes' if not args['eps_smoothing'] else 'no',
            'nc': args['char_order'],
            'nw': args['word_order'],
            'space': 'yes' if args['whitespace'] else 'no',
        })


class CHRFScore(Score):
    """A convenience class to represent chrF scores.

    :param score: The chrF (chrF++) score.
    :param char_order: The character n-gram order.
    :param word_order: The word n-gram order. If equals to 2, the metric is referred to as chrF++.
    :param beta: Determine the importance of recall w.r.t precision.
    """
    def __init__(self, score: float, char_order: int, word_order: int, beta: int):
        """`CHRFScore` initializer."""
        self.beta = beta
        self.char_order = char_order
        self.word_order = word_order

        # Add + signs to denote chrF+ variant
        name = f'chrF{self.beta}' + '+' * self.word_order

        super().__init__(name, score)


class CHRF(Metric):
    """Computes the chrF(++) metric given hypotheses and references.

    :param char_order: Character n-gram order.
    :param word_order: Word n-gram order. If equals to 2, the metric is referred to as chrF++.
    :param beta: Determine the importance of recall w.r.t precision.
    :param lowercase: Enable case-insensitivity.
    :param whitespace: If `True`, include whitespaces when extracting character n-grams.
    :param eps_smoothing: If `True`, applies epsilon smoothing similar
    to reference chrF++.py, NLTK and Moses implementations. Otherwise,
    it takes into account effective match order similar to sacreBLEU < 2.0.0.
    :param references: A sequence of reference documents with document being
    defined as a sequence of reference strings. If given, the reference n-grams
    will be pre-computed and cached for faster re-computation across many systems.
    """

    # Maximum character n-gram order to take into account
    CHAR_ORDER = 6

    # chrF+ additionally takes into account some of the word n-grams
    WORD_ORDER = 0

    # Defaults to 2 (per http://www.aclweb.org/anthology/W16-2341)
    BETA = 2

    # Cache string.punctuation for chrF+' punctuation stripper
    _PUNCTS = set('!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~')

    _SIGNATURE_TYPE = CHRFSignature

    def __init__(self, char_order: int = CHAR_ORDER,
                 word_order: int = WORD_ORDER,
                 beta: int = BETA,
                 lowercase: bool = False,
                 whitespace: bool = False,
                 eps_smoothing: bool = False,
                 references: Optional[Sequence[Sequence[str]]] = None):
        """`CHRF` initializer."""
        super().__init__()

        self.beta = beta
        self.char_order = char_order
        self.word_order = word_order
        self.order = self.char_order + self.word_order
        self.lowercase = lowercase
        self.whitespace = whitespace
        self.eps_smoothing = eps_smoothing

        if references is not None:
            # Pre-compute reference ngrams
            self._ref_cache = self._cache_references(references)

    @staticmethod
    def _get_match_statistics(hyp_ngrams: Counter, ref_ngrams: Counter) -> List[int]:
        """Computes the match statistics between hypothesis and reference n-grams.

        :param hyp_ngrams: A `Counter` holding hypothesis n-grams.
        :param ref_ngrams: A `Counter` holding reference n-grams.
        :return: A list of three numbers denoting hypothesis n-gram count,
            reference n-gram count and the intersection count.
        """
        # Counter's internal intersection is not that fast, count manually
        match_count, hyp_count = 0, 0
        for ng, count in hyp_ngrams.items():
            hyp_count += count
            if ng in ref_ngrams:
                match_count += min(count, ref_ngrams[ng])

        return [
            # Don't count hits if no reference exists for that n-gram
            hyp_count if ref_ngrams else 0,
            sum(ref_ngrams.values()),
            match_count,
        ]

    def _remove_punctuation(self, sent: str) -> List[str]:
        """Separates out punctuations from beginning and end of words for chrF.
        Adapted from https://github.com/m-popovic/chrF

        :param sent: A string.
        :return: A list of words.
        """
        tokenized = []
        for w in sent.split():
            if len(w) == 1:
                tokenized.append(w)
            else:
                # NOTE: This splits '(hi)' to '(hi' and ')' (issue #124)
                if w[-1] in self._PUNCTS:
                    tokenized += [w[:-1], w[-1]]
                elif w[0] in self._PUNCTS:
                    tokenized += [w[0], w[1:]]
                else:
                    tokenized.append(w)
        return tokenized

    def _preprocess_segment(self, sent: str) -> str:
        """Given a sentence, apply optional lowercasing.

        :param sent: The input sentence string.
        :return: The pre-processed output string.
        """
        return sent.lower() if self.lowercase else sent

    def _compute_f_score(self, statistics: List[int]) -> float:
        """Compute the chrF score given the n-gram match statistics.

        :param statistics: A flattened list of 3 * (`char_order` + `word_order`)
            elements giving the [hyp, ref, match] counts for each order.
        :return: The final f_beta score between [0, 100].
        """
        eps = 1e-16
        score = 0.0
        effective_order = 0
        factor = self.beta ** 2
        avg_prec, avg_rec = 0.0, 0.0

        for i in range(self.order):
            n_hyp, n_ref, n_match = statistics[3 * i: 3 * i + 3]

            # chrF++.py style EPS smoothing (also used by Moses and NLTK)
            prec = n_match / n_hyp if n_hyp > 0 else eps
            rec = n_match / n_ref if n_ref > 0 else eps

            denom = factor * prec + rec
            score += ((1 + factor) * prec * rec / denom) if denom > 0 else eps

            # sacreBLEU <2.0.0 style effective order smoothing
            if n_hyp > 0 and n_ref > 0:
                avg_prec += prec
                avg_rec += rec
                effective_order += 1

        if self.eps_smoothing:
            return 100 * score / self.order

        if effective_order == 0:
            avg_prec = avg_rec = 0.0
        else:
            avg_prec /= effective_order
            avg_rec /= effective_order

        if avg_prec + avg_rec:
            score = (1 + factor) * avg_prec * avg_rec
            score /= ((factor * avg_prec) + avg_rec)
            return 100 * score
        else:
            return 0.0

    def _compute_score_from_stats(self, stats: List[int]) -> CHRFScore:
        """Computes the final score from already aggregated statistics.

        :param stats: A list or numpy array of segment-level statistics.
        :return: A `CHRFScore` object.
        """
        return CHRFScore(
            self._compute_f_score(stats), self.char_order,
            self.word_order, self.beta)

    def _aggregate_and_compute(self, stats: List[List[int]]) -> CHRFScore:
        """Computes the final score given the pre-computed corpus statistics.

        :param stats: A list of segment-level statistics
        :return: A `CHRFScore` object.
        """
        return self._compute_score_from_stats(sum_of_lists(stats))

    def _extract_reference_info(self, refs: Sequence[str]) -> Dict[str, List[List[Counter]]]:
        """Given a list of reference segments, extract the character and word n-grams.

        :param refs: A sequence of reference segments.
        :return: A list where each element contains n-grams per reference segment.
        """
        ngrams = []

        for ref in refs:
            # extract character n-grams
            stats = extract_all_char_ngrams(ref, self.char_order, self.whitespace)

            # Check chrF+ mode
            if self.word_order > 0:
                ref_words = self._remove_punctuation(ref)

                for n in range(self.word_order):
                    stats.append(extract_word_ngrams(ref_words, n + 1))

            ngrams.append(stats)

        return {'ref_ngrams': ngrams}

    def _compute_segment_statistics(
            self, hypothesis: str, ref_kwargs: Dict) -> List[int]:
        """Given a (pre-processed) hypothesis sentence and already computed
        reference n-grams, returns the best match statistics across the
        references.

        :param hypothesis: Hypothesis sentence.
        :param ref_kwargs: A dictionary with key `ref_ngrams` which is a list
        where each sublist contains n-gram counters for a particular reference sentence.
        :return: A list of integers where each triplet denotes [hyp, ref, match]
        statistics.
        """
        best_stats = []
        best_f_score = -1.0

        # extract character n-grams
        all_hyp_ngrams = extract_all_char_ngrams(
            hypothesis, self.char_order, self.whitespace)

        # Check chrF+ mode to see if we'll add word n-grams as well
        if self.word_order > 0:
            # Primitive tokenization: separate out punctuations
            hwords = self._remove_punctuation(hypothesis)
            _range = range(1, self.word_order + 1)
            all_hyp_ngrams.extend([extract_word_ngrams(hwords, n) for n in _range])

        # Iterate over multiple references, pick the one with best F score
        for _ref_ngrams in ref_kwargs['ref_ngrams']:
            stats = []
            # Traverse all orders
            for h, r in zip(all_hyp_ngrams, _ref_ngrams):
                stats.extend(self._get_match_statistics(h, r))
            f_score = self._compute_f_score(stats)

            if f_score > best_f_score:
                best_f_score = f_score
                best_stats = stats

        return best_stats



import pytest

from sacrebleu.tokenizers.tokenizer_ter import TercomTokenizer


test_cases_default = [
    ("a b c d", "a b c d"),
    ("", ""),
    ("a b c d.", "a b c d."),
    ("A B C D.", "a b c d."),
]

test_cases_no_punct = [
    ("a b c d.", "a b c d"),
    ("A ; B ) C : D.", "a b c d"),
]

test_cases_norm = [
    ("a b (c) d.", "a b ( c ) d ."),
    ("Jim's car.", "jim 's car ."),
    ("4.2", "4.2"),
]

test_cases_asian = [
    ("美众院公布对特", "美 众 院 公 布 对 特"),  # Chinese
    ("りの拳銃を持", "りの 拳 銃 を 持"),  # Japanese, first two letters are Hiragana
]


@pytest.mark.parametrize("input, expected", test_cases_default)
def test_ter_tokenizer_default(input, expected):
    tokenizer = TercomTokenizer()
    assert tokenizer(input) == expected


@pytest.mark.parametrize("input, expected", test_cases_no_punct)
def test_ter_tokenizer_nopunct(input, expected):
    tokenizer = TercomTokenizer(no_punct=True)
    assert tokenizer(input) == expected


@pytest.mark.parametrize("input, expected", test_cases_norm)
def test_ter_tokenizer_norm(input, expected):
    tokenizer = TercomTokenizer(normalized=True)
    assert tokenizer(input) == expected


@pytest.mark.parametrize("input, expected", test_cases_asian)
def test_ter_tokenizer_asian(input, expected):
    tokenizer = TercomTokenizer(normalized=True, asian_support=True)
    assert tokenizer(input) == expected
